@@ -22,6 +22,7 @@ const Gettext = imports.gettext.domain('gnome-shell');
 const _ = Gettext.gettext;
 
 const NMConnectionCategory = {
+    INVALID: 'invalid',
     WIRED: 'wired',
     WIRELESS: 'wireless',
     WWAN: 'wwan',
@@ -92,16 +93,31 @@ function NMNetworkMenuItem() {
 }
 
 NMNetworkMenuItem.prototype = {
-    __proto__: PopupMenu.PopupImageMenuItem.prototype,
+    __proto__: PopupMenu.PopupBaseMenuItem.prototype,
 
     _init: function(accessPoints, title, params) {
+        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, params);
+
         accessPoints = sortAccessPoints(accessPoints);
         this.bestAP = accessPoints[0];
 
         let ssid = this.bestAP.get_ssid();
         title = title || NetworkManager.utils_ssid_to_utf8(ssid) || _("<unknown>");
 
-        PopupMenu.PopupImageMenuItem.prototype._init.call(this, title, this._getIcon(), params);
+        this._label = new St.Label({ text: title });
+        this.addActor(this._label);
+        this._icons = new St.BoxLayout({ style_class: 'nm-menu-item-icons' });
+        this.addActor(this._icons, { align: St.Align.END });
+
+        this._signalIcon = new St.Icon({ icon_name: this._getIcon(),
+                                         style_class: 'popup-menu-icon' });
+        this._icons.add_actor(this._signalIcon);
+
+        this._secureIcon = new St.Icon({ style_class: 'popup-menu-icon' });
+        if (this.bestAP._secType != NMAccessPointSecurity.UNKNOWN &&
+            this.bestAP._secType != NMAccessPointSecurity.NONE)
+            this._secureIcon.icon_name = 'network-wireless-encrypted';
+        this._icons.add_actor(this._secureIcon);
 
         this._accessPoints = [ ];
         for (let i = 0; i < accessPoints.length; i++) {
@@ -120,7 +136,7 @@ NMNetworkMenuItem.prototype = {
         if (strength > this.bestAP.strength)
             this.bestAP = ap;
 
-        this.setIcon(this._getIcon());
+        this._signalIcon.icon_name = this._getIcon();
     },
 
     _getIcon: function() {
@@ -238,7 +254,14 @@ NMWiredSectionTitleMenuItem.prototype = {
             return;
         }
 
-        if (this._switch.state)
+        let newState = this._switch.state;
+
+        // Immediately reset the switch to false, it will be updated appropriately
+        // by state-changed signals in devices (but fixes the VPN not being in sync
+        // if the ActiveConnection object is never seen by libnm-glib)
+        this._switch.setToggleState(false);
+
+        if (newState)
             this._device.activate();
         else
             this._device.deactivate();
@@ -379,7 +402,7 @@ NMDevice.prototype = {
         // pick the most recently used connection and connect to that
         // or if no connections ever set, create an automatic one
         if (this._connections.length > 0) {
-            this._client.activate_connection(this._connections[0].connection.path, this.device, null, null);
+            this._client.activate_connection(this._connections[0].connection, this.device, null, null);
         } else if (this._autoConnectionName) {
             let connection = this._createAutomaticConnection();
             this._client.add_and_activate_connection(connection, this.device, null, null);
@@ -408,7 +431,7 @@ NMDevice.prototype = {
     },
 
     checkConnection: function(connection) {
-        let exists = this._findConnection(connection) != -1;
+        let exists = this._findConnection(connection._uuid) != -1;
         let valid = this.connectionValid(connection);
         if (exists && !valid)
             this.removeConnection(connection);
@@ -449,8 +472,10 @@ NMDevice.prototype = {
             obj.item.destroy();
         this._connections.splice(pos, 1);
 
-        if (this._connections.length == 0) {
+        if (this._connections.length <= 1) {
             // We need to show the automatic connection again
+            // (or in the case of NMDeviceWired, we want to hide
+            // the only explicit connection)
             this._clearSection();
             this._createSection();
         }
@@ -478,15 +503,27 @@ NMDevice.prototype = {
             /* Translators: this is for network connections that require some kind of key or password */
             return _("authentication required");
         case NetworkManager.DeviceState.UNAVAILABLE:
-            // we don't check if the carrier property is actually false, as that causes race
-            // conditions if state is changed before the new carrier value is picked by libnm-glib
-            if (this.device.capabilities & NetworkManager.DeviceCapabilities.CARRIER_DETECT)
-                /* Translators: this is for wired network devices that are physically disconnected */
-                return _("cable unplugged");
-            else
-                /* Translators: this is for a network device that cannot be activated (for example it
-                   is disabled by rfkill, or it has no coverage */
-                return _("unavailable");
+            // This state is actually a compound of various states (generically unavailable,
+            // firmware missing, carrier not available), that are exposed by different properties
+            // (whose state may or may not updated when we receive state-changed).
+            if (!this._firmwareMissingId)
+                this._firmwareMissingId = this.device.connect('notify::firmware-missing', Lang.bind(this, this._substateChanged));
+            if (this.device.firmware_missing) {
+                /* Translators: this is for devices that require some kind of firmware or kernel
+                   module, which is missing */
+                return _("firmware missing");
+            }
+            if (this.device.capabilities & NetworkManager.DeviceCapabilities.CARRIER_DETECT) {
+                if (!this._carrierChangedId)
+                    this._carrierChangedId = this.device.connect('notify::carrier', Lang.bind(this, this._substateChanged));
+                if (!this.carrier) {
+                    /* Translators: this is for wired network devices that are physically disconnected */
+                    return _("cable unplugged");
+                }
+            }
+            /* Translators: this is for a network device that cannot be activated (for example it
+               is disabled by rfkill, or it has no coverage */
+            return _("unavailable");
         case NetworkManager.DeviceState.FAILED:
             return _("connection failed");
         default:
@@ -520,8 +557,7 @@ NMDevice.prototype = {
     },
 
     _shouldShowConnectionList: function() {
-        return (this.device.state == NetworkManager.DeviceState.DISCONNECTED ||
-                this.device.state == NetworkManager.DeviceState.ACTIVATED);
+        return (this.device.state >= NetworkManager.DeviceState.DISCONNECTED);
     },
 
     _createSection: function() {
@@ -552,10 +588,11 @@ NMDevice.prototype = {
     },
 
     _createConnectionItem: function(obj) {
-        let path = obj.connection.path;
+        let connection = obj.connection;
         let item = new PopupMenu.PopupMenuItem(obj.name);
+
         item.connect('activate', Lang.bind(this, function() {
-            this._client.activate_connection(path, this.device, null, null);
+            this._client.activate_connection(connection, this.device, null, null);
         }));
         return item;
     },
@@ -595,6 +632,16 @@ NMDevice.prototype = {
             break;
         }
 
+        if (this._carrierChangedId) {
+            // see above for why this is needed
+            GObject.Object.prototype.disconnect.call(this.device, this._carrierChangedId);
+            this._carrierChangedId = 0;
+        }
+        if (this._firmwareChangedId) {
+            GObject.Object.prototype.disconnect.call(this.device, this._firmwareChangedId);
+            this._firmwareChangedId = 0;
+        }
+
         this.statusItem.setStatus(this.getStatusLabel());
         this.statusItem.setToggleState(this.connected);
 
@@ -603,6 +650,12 @@ NMDevice.prototype = {
         this.emit('state-changed');
     },
 
+    _substateChanged: function() {
+        this.statusItem.setStatus(this.getStatusLabel());
+
+        this.emit('state-changed');
+    },
+
     _getDescription: function() {
         let dev_product = this.device.get_product();
         let dev_vendor = this.device.get_vendor();
@@ -650,6 +703,20 @@ NMDeviceWired.prototype = {
         return true;
     },
 
+    _createSection: function() {
+        NMDevice.prototype._createSection.call(this);
+
+        // if we have only one connection (normal or automatic)
+        // we hide the connection list, and use the switch to control
+        // the device
+        // we can do it here because addConnection and removeConnection
+        // both call _createSection at some point
+        if (this._connections.length <= 1)
+            this.section.actor.hide();
+        else
+            this.section.actor.show();
+    },
+
     _createAutomaticConnection: function() {
         let connection = new NetworkManager.Connection();
         connection._uuid = NetworkManager.utils_uuid_generate();
@@ -879,7 +946,13 @@ NMDeviceVPN.prototype = {
     },
 
     get connected() {
-        return true;
+        return !!this._activeConnection;
+    },
+
+    setActiveConnection: function(activeConnection) {
+        NMDevice.prototype.setActiveConnection.call(this, activeConnection);
+
+        this.emit('active-connection-changed');
     },
 
     _shouldShowConnectionList: function() {
@@ -1001,7 +1074,7 @@ NMDeviceWireless.prototype = {
             for (let i = 0; i < bestApObj.accessPoints.length; i++) {
                 let ap = bestApObj.accessPoints[i];
                 if (this._connectionValidForAP(best, ap)) {
-                    this._client.activate_connection(best.path, this.device, ap.dbus_path, null);
+                    this._client.activate_connection(best, this.device, ap.dbus_path, null);
                     break;
                 }
             }
@@ -1136,7 +1209,7 @@ NMDeviceWireless.prototype = {
             let accessPoints = sortAccessPoints(accessPointObj.accessPoints);
             for (let i = 0; i < accessPoints.length; i++) {
                 if (this._connectionValidForAP(connection, accessPoints[i])) {
-                    this._client.activate_connection(connection.path, this.device, accessPoints[i].dbus_path, null);
+                    this._client.activate_connection(connection, this.device, accessPoints[i].dbus_path, null);
                     break;
                 }
             }
@@ -1478,6 +1551,9 @@ NMApplet.prototype = {
             device: new NMDeviceVPN(this._client),
             item: new NMWiredSectionTitleMenuItem(_("VPN Connections"))
         };
+        this._devices.vpn.device.connect('active-connection-changed', Lang.bind(this, function() {
+            this._devices.vpn.item.updateForDevice(this._devices.vpn.device);
+        }));
         this._devices.vpn.item.updateForDevice(this._devices.vpn.device);
         this._devices.vpn.section.addMenuItem(this._devices.vpn.item);
         this._devices.vpn.section.addMenuItem(this._devices.vpn.device.section);
@@ -1732,14 +1808,9 @@ NMApplet.prototype = {
             if (a.default6)
                 default_ip6 = a;
 
-            if (a.state == NetworkManager.ActiveConnectionState.ACTIVATING) {
+            if (a.state == NetworkManager.ActiveConnectionState.ACTIVATING)
                 activating = a;
 
-                // don't set activating connections to devices, NMDevice:state-changed
-                // should take care of rebuilding the menu
-                continue;
-            }
-
             if (!a._primaryDevice) {
                 if (a._type != NetworkManager.SETTING_VPN_SETTING_NAME) {
                     // find a good device to be considered primary
@@ -1753,7 +1824,7 @@ NMApplet.prototype = {
                         }
                     }
                 } else
-                    a._primaryDevice = this._vpnDevice;
+                    a._primaryDevice = this._devices.vpn.device
 
                 if (a._primaryDevice)
                     a._primaryDevice.setActiveConnection(a);
@@ -1857,11 +1928,12 @@ NMApplet.prototype = {
             this._connections.splice(connection);
 
         let section = connection._section;
+
         if (section == NMConnectionCategory.VPN) {
             this._devices.vpn.device.removeConnection(connection);
             if (this._devices.vpn.device.empty)
                 this._devices.vpn.section.actor.hide();
-        } else {
+        } else if (section != NMConnectionCategory.INVALID) {
             let devices = this._devices[section].devices;
             for (let i = 0; i < devices.length; i++)
                 devices[i].removeConnection(connection);
@@ -1875,16 +1947,19 @@ NMApplet.prototype = {
     _updateConnection: function(connection) {
         let connectionSettings = connection.get_setting_by_name(NetworkManager.SETTING_CONNECTION_SETTING_NAME);
         connection._type = connectionSettings.type;
-        connection._section = this._ctypes[connection._type];
+
+        connection._section = this._ctypes[connection._type] || NMConnectionCategory.INVALID;
         connection._name = connectionSettings.id;
         connection._uuid = connectionSettings.uuid;
         connection._timestamp = connectionSettings.timestamp;
 
         let section = connection._section;
+
+        if (connection._section == NMConnectionCategory.INVALID)
+            return;
         if (section == NMConnectionCategory.VPN) {
             this._devices.vpn.device.checkConnection(connection);
             this._devices.vpn.section.actor.show();
-            connection._everAdded = true;
         } else {
             let devices = this._devices[section].devices;
             for (let i = 0; i < devices.length; i++) {
